@@ -4,8 +4,8 @@ module Bundler
   # Represents metadata from when the Bundler gem was built.
   module BuildMetadata
     # begin ivars
-    @built_at = "2018-09-18".freeze
-    @git_commit_sha = "1da10a61b".freeze
+    @built_at = "2019-01-04".freeze
+    @git_commit_sha = "d7ad2192f".freeze
     @release = true
     # end ivars
 
@@ -30,12 +30,19 @@ module Bundler
       # If Bundler has been installed without its .git directory and without a
       # commit instance variable then we can't determine its commits SHA.
       git_dir = File.join(File.expand_path("../../..", __FILE__), ".git")
-      return "unknown" unless File.directory?(git_dir)
-
-      # Otherwise shell out to git.
-      @git_commit_sha = Dir.chdir(File.expand_path("..", __FILE__)) do
-        `git rev-parse --short HEAD`.strip.freeze
+      if File.directory?(git_dir)
+        return @git_commit_sha = Dir.chdir(git_dir) { `git rev-parse --short HEAD`.strip.freeze }
       end
+
+      # If Bundler is a submodule in RubyGems, get the submodule commit
+      git_sub_dir = File.join(File.expand_path("../../../..", __FILE__), ".git")
+      if File.directory?(git_sub_dir)
+        return @git_commit_sha = Dir.chdir(git_sub_dir) do
+          `git ls-tree --abbrev=8 HEAD bundler`.split(/\s/).fetch(2, "").strip.freeze
+        end
+      end
+
+      @git_commit_sha ||= "unknown"
     end
 
     # Whether this is an official release build of Bundler.
