@@ -1,4 +1,5 @@
 # frozen_string_literal: true
+
 require "monitor"
 require "rubygems"
 require "rubygems/config_file"
@@ -74,12 +75,33 @@ module Bundler
     def spec_missing_extensions?(spec, default = true)
       return spec.missing_extensions? if spec.respond_to?(:missing_extensions?)
 
-      return false if spec.respond_to?(:default_gem?) && spec.default_gem?
+      return false if spec_default_gem?(spec)
       return false if spec.extensions.empty?
 
       default
     end
 
+    def spec_default_gem?(spec)
+      spec.respond_to?(:default_gem?) && spec.default_gem?
+    end
+
+    def spec_matches_for_glob(spec, glob)
+      return spec.matches_for_glob(glob) if spec.respond_to?(:matches_for_glob)
+
+      spec.load_paths.map do |lp|
+        Dir["#{lp}/#{glob}#{suffix_pattern}"]
+      end.flatten(1)
+    end
+
+    def spec_extension_dir(spec)
+      return unless spec.respond_to?(:extension_dir)
+      spec.extension_dir
+    end
+
+    def stub_set_spec(stub, spec)
+      stub.instance_variable_set(:@spec, spec)
+    end
+
     def path(obj)
       obj.to_s
     end
@@ -110,7 +132,11 @@ module Bundler
     end
 
     def inflate(obj)
-      Gem.inflate(obj)
+      if defined?(Gem::Util)
+        Gem::Util.inflate(obj)
+      else
+        Gem.inflate(obj)
+      end
     end
 
     def sources=(val)
@@ -150,6 +176,10 @@ module Bundler
       Gem.post_reset_hooks
     end
 
+    def suffix_pattern
+      Gem.suffix_pattern
+    end
+
     def gem_cache
       gem_path.map {|p| File.expand_path("cache", p) }
     end
@@ -157,7 +187,7 @@ module Bundler
     def spec_cache_dirs
       @spec_cache_dirs ||= begin
         dirs = gem_path.map {|dir| File.join(dir, "specifications") }
-        dirs << Gem.spec_cache_dir if Gem.respond_to?(:spec_cache_dir) # Not in Rubygems 2.0.3 or earlier
+        dirs << Gem.spec_cache_dir if Gem.respond_to?(:spec_cache_dir) # Not in RubyGems 2.0.3 or earlier
         dirs.uniq.select {|dir| File.directory? dir }
       end
     end
@@ -171,7 +201,7 @@ module Bundler
     end
 
     def repository_subdirectories
-      %w(cache doc gems specifications)
+      %w[cache doc gems specifications]
     end
 
     def clear_paths
@@ -182,8 +212,12 @@ module Bundler
       Gem.bin_path(gem, bin, ver)
     end
 
+    def path_separator
+      File::PATH_SEPARATOR
+    end
+
     def preserve_paths
-      # this is a no-op outside of Rubygems 1.8
+      # this is a no-op outside of RubyGems 1.8
       yield
     end
 
@@ -204,6 +238,10 @@ module Bundler
       Gem.load_plugins if Gem.respond_to?(:load_plugins)
     end
 
+    def load_plugin_files(files)
+      Gem.load_plugin_files(files) if Gem.respond_to?(:load_plugin_files)
+    end
+
     def ui=(obj)
       Gem::DefaultUserInteraction.ui = obj
     end
@@ -225,9 +263,9 @@ module Bundler
       {} # if we can't download them, there aren't any
     end
 
-    # TODO: This is for older versions of Rubygems... should we support the
+    # TODO: This is for older versions of RubyGems... should we support the
     # X-Gemfile-Source header on these old versions?
-    # Maybe the newer implementation will work on older Rubygems?
+    # Maybe the newer implementation will work on older RubyGems?
     # It seems difficult to keep this implementation and still send the header.
     def fetch_all_remote_specs(remote)
       old_sources = Bundler.rubygems.sources
@@ -265,6 +303,7 @@ module Bundler
 
     def spec_from_gem(path, policy = nil)
       require "rubygems/security"
+      require "bundler/psyched_yaml"
       gem_from_path(path, security_policies[policy]).spec
     rescue Gem::Package::FormatError
       raise GemspecError, "Could not read gem at #{path}. It may be corrupted."
@@ -292,13 +331,13 @@ module Bundler
     def download_gem(spec, uri, path)
       uri = Bundler.settings.mirror_for(uri)
       fetcher = Gem::RemoteFetcher.new(configuration[:http_proxy])
-      Bundler::Retry.new("download gem #{uri}", Gem::RemoteFetcher::FetchError).attempts do
+      Bundler::Retry.new("download gem from #{uri}").attempts do
         fetcher.download(spec, uri, path)
       end
     end
 
     def security_policy_keys
-      %w(High Medium Low AlmostNo No).map {|level| "#{level}Security" }
+      %w[High Medium Low AlmostNo No].map {|level| "#{level}Security" }
     end
 
     def security_policies
@@ -320,53 +359,57 @@ module Bundler
       end
     end
 
-    def replace_gem(specs)
+    def binstubs_call_gem?
+      true
+    end
+
+    def stubs_provide_full_functionality?
+      false
+    end
+
+    def replace_gem(specs, specs_by_name)
       reverse_rubygems_kernel_mixin
 
-      executables = specs.map(&:executables).flatten
+      executables = nil
 
       kernel = (class << ::Kernel; self; end)
       [kernel, ::Kernel].each do |kernel_class|
         redefine_method(kernel_class, :gem) do |dep, *reqs|
-          if executables.include? File.basename(caller.first.split(":").first)
+          executables ||= specs.map(&:executables).flatten if ::Bundler.rubygems.binstubs_call_gem?
+          if executables && executables.include?(File.basename(caller.first.split(":").first))
             break
           end
+
           reqs.pop if reqs.last.is_a?(Hash)
 
           unless dep.respond_to?(:name) && dep.respond_to?(:requirement)
             dep = Gem::Dependency.new(dep, reqs)
           end
 
-          spec = specs.find {|s| s.name == dep.name }
-
-          if spec.nil?
-
-            e = Gem::LoadError.new "#{dep.name} is not part of the bundle. Add it to Gemfile."
-            e.name = dep.name
-            if e.respond_to?(:requirement=)
-              e.requirement = dep.requirement
-            else
-              e.version_requirement = dep.requirement
-            end
-            raise e
-          elsif dep !~ spec
-            e = Gem::LoadError.new "can't activate #{dep}, already activated #{spec.full_name}. " \
-                                   "Make sure all dependencies are added to Gemfile."
-            e.name = dep.name
-            if e.respond_to?(:requirement=)
-              e.requirement = dep.requirement
-            else
-              e.version_requirement = dep.requirement
-            end
-            raise e
+          if spec = specs_by_name[dep.name]
+            return true if dep.matches_spec?(spec)
           end
 
-          true
+          message = if spec.nil?
+            "#{dep.name} is not part of the bundle." \
+            " Add it to your #{Bundler.default_gemfile.basename}."
+          else
+            "can't activate #{dep}, already activated #{spec.full_name}. " \
+            "Make sure all dependencies are added to Gemfile."
+          end
+
+          e = Gem::LoadError.new(message)
+          e.name = dep.name
+          if e.respond_to?(:requirement=)
+            e.requirement = dep.requirement
+          elsif e.respond_to?(:version_requirement=)
+            e.version_requirement = dep.requirement
+          end
+          raise e
         end
 
-        # TODO: delete this in 2.0, it's a backwards compatibility shim
-        # see https://github.com/bundler/bundler/issues/5102
-        kernel_class.send(:public, :gem)
+        # backwards compatibility shim, see https://github.com/bundler/bundler/issues/5102
+        kernel_class.send(:public, :gem) if Bundler.feature_flag.setup_makes_kernel_gem_public?
       end
     end
 
@@ -393,23 +436,37 @@ module Bundler
     # Used to make bin stubs that are not created by bundler work
     # under bundler. The new Gem.bin_path only considers gems in
     # +specs+
-    def replace_bin_path(specs)
+    def replace_bin_path(specs, specs_by_name)
       gem_class = (class << Gem; self; end)
 
       redefine_method(gem_class, :find_spec_for_exe) do |gem_name, *args|
         exec_name = args.first
 
+        spec_with_name = specs_by_name[gem_name]
         spec = if exec_name
-          specs.find {|s| s.name == gem_name && s.executables.include?(exec_name) } ||
+          if spec_with_name && spec_with_name.executables.include?(exec_name)
+            spec_with_name
+          else
             specs.find {|s| s.executables.include?(exec_name) }
+          end
         else
-          specs.find {|s| s.name == gem_name }
+          spec_with_name
         end
-        raise(Gem::Exception, "can't find executable #{exec_name}") unless spec
+
+        unless spec
+          message = "can't find executable #{exec_name} for gem #{gem_name}"
+          if !exec_name || spec_with_name.nil?
+            message += ". #{gem_name} is not currently included in the bundle, " \
+                       "perhaps you meant to add it to your #{Bundler.default_gemfile.basename}?"
+          end
+          raise Gem::Exception, message
+        end
+
         raise Gem::Exception, "no default executable for #{spec.full_name}" unless exec_name ||= spec.default_executable
-        unless spec.name == name
-          Bundler::SharedHelpers.major_deprecation \
-            "Bundler is using a binstub that was created for a different gem.\n" \
+
+        unless spec.name == gem_name
+          Bundler::SharedHelpers.major_deprecation 3,
+            "Bundler is using a binstub that was created for a different gem (#{spec.name}).\n" \
             "You should run `bundle binstub #{gem_name}` " \
             "to work around a system/bundle conflict."
         end
@@ -423,8 +480,10 @@ module Bundler
         # Copy of Rubygems activate_bin_path impl
         requirement = args.last
         spec = find_spec_for_exe name, exec_name, [requirement]
-        Gem::LOADED_SPECS_MUTEX.synchronize { spec.activate }
-        spec.bin_file exec_name
+
+        gem_bin = File.join(spec.full_gem_path, spec.bindir, exec_name)
+        gem_from_path_bin = File.join(File.dirname(spec.loaded_from), spec.bindir, exec_name)
+        File.exist?(gem_bin) ? gem_bin : gem_from_path_bin
       end
 
       redefine_method(gem_class, :bin_path) do |name, *args|
@@ -447,21 +506,24 @@ module Bundler
       redefine_method(gem_class, :refresh) {}
     end
 
-    # Replace or hook into Rubygems to provide a bundlerized view
+    # Replace or hook into RubyGems to provide a bundlerized view
     # of the world.
     def replace_entrypoints(specs)
-      replace_gem(specs)
+      specs_by_name = specs.reduce({}) do |h, s|
+        h[s.name] = s
+        h
+      end
 
+      replace_gem(specs, specs_by_name)
       stub_rubygems(specs)
-
-      replace_bin_path(specs)
+      replace_bin_path(specs, specs_by_name)
       replace_refresh
 
       Gem.clear_paths
     end
 
-    # This backports the correct segment generation code from Rubygems 1.4+
-    # by monkeypatching it into the method in Rubygems 1.3.6 and 1.3.7.
+    # This backports the correct segment generation code from RubyGems 1.4+
+    # by monkeypatching it into the method in RubyGems 1.3.6 and 1.3.7.
     def backport_segment_generation
       redefine_method(Gem::Version, :segments) do
         @segments ||= @version.scan(/[0-9]+|[a-z]+/i).map do |s|
@@ -480,7 +542,7 @@ module Bundler
     end
 
     # This backports base_dir which replaces installation path
-    # Rubygems 1.8+
+    # RubyGems 1.8+
     def backport_base_dir
       redefine_method(Gem::Specification, :base_dir) do
         return Gem.dir unless loaded_from
@@ -512,8 +574,10 @@ module Bundler
       @replaced_methods.each do |(sym, klass), method|
         redefine_method(klass, sym, method)
       end
-      post_reset_hooks.reject! do |proc|
-        proc.binding.eval("__FILE__") == __FILE__
+      if Binding.public_method_defined?(:source_location)
+        post_reset_hooks.reject! {|proc| proc.binding.source_location[0] == __FILE__ }
+      else
+        post_reset_hooks.reject! {|proc| proc.binding.eval("__FILE__") == __FILE__ }
       end
       @replaced_methods.clear
     end
@@ -549,7 +613,7 @@ module Bundler
       end
     end
 
-    # Rubygems 1.4 through 1.6
+    # RubyGems 1.4 through 1.6
     class Legacy < RubygemsIntegration
       def initialize
         super
@@ -560,7 +624,7 @@ module Bundler
       end
 
       def stub_rubygems(specs)
-        # Rubygems versions lower than 1.7 use SourceIndex#from_gems_in
+        # RubyGems versions lower than 1.7 use SourceIndex#from_gems_in
         source_index_class = (class << Gem::SourceIndex; self; end)
         redefine_method(source_index_class, :from_gems_in) do |*args|
           Gem::SourceIndex.new.tap do |source_index|
@@ -592,7 +656,7 @@ module Bundler
       end
     end
 
-    # Rubygems versions 1.3.6 and 1.3.7
+    # RubyGems versions 1.3.6 and 1.3.7
     class Ancient < Legacy
       def initialize
         super
@@ -600,7 +664,7 @@ module Bundler
       end
     end
 
-    # Rubygems 1.7
+    # RubyGems 1.7
     class Transitional < Legacy
       def stub_rubygems(specs)
         stub_source_index(specs)
@@ -614,7 +678,7 @@ module Bundler
       end
     end
 
-    # Rubygems 1.8.5-1.8.19
+    # RubyGems 1.8.5-1.8.19
     class Modern < RubygemsIntegration
       def stub_rubygems(specs)
         Gem::Specification.all = specs
@@ -635,9 +699,9 @@ module Bundler
       end
     end
 
-    # Rubygems 1.8.0 to 1.8.4
+    # RubyGems 1.8.0 to 1.8.4
     class AlmostModern < Modern
-      # Rubygems [>= 1.8.0, < 1.8.5] has a bug that changes Gem.dir whenever
+      # RubyGems [>= 1.8.0, < 1.8.5] has a bug that changes Gem.dir whenever
       # you call Gem::Installer#install with an :install_dir set. We have to
       # change it back for our sudo mode to work.
       def preserve_paths
@@ -648,9 +712,9 @@ module Bundler
       end
     end
 
-    # Rubygems 1.8.20+
+    # RubyGems 1.8.20+
     class MoreModern < Modern
-      # Rubygems 1.8.20 and adds the skip_validation parameter, so that's
+      # RubyGems 1.8.20 and adds the skip_validation parameter, so that's
       # when we start passing it through.
       def build(spec, skip_validation = false)
         require "rubygems/builder"
@@ -658,7 +722,7 @@ module Bundler
       end
     end
 
-    # Rubygems 2.0
+    # RubyGems 2.0
     class Future < RubygemsIntegration
       def stub_rubygems(specs)
         Gem::Specification.all = specs
@@ -666,6 +730,10 @@ module Bundler
         Gem.post_reset do
           Gem::Specification.all = specs
         end
+
+        redefine_method((class << Gem; self; end), :finish_resolve) do |*|
+          []
+        end
       end
 
       def all_specs
@@ -700,7 +768,9 @@ module Bundler
         uri = Bundler.settings.mirror_for(uri)
         fetcher = gem_remote_fetcher
         fetcher.headers = { "X-Gemfile-Source" => spec.remote.original_uri.to_s } if spec.remote.original_uri
-        fetcher.download(spec, uri, path)
+        Bundler::Retry.new("download gem from #{uri}").attempts do
+          fetcher.download(spec, uri, path)
+        end
       end
 
       def gem_remote_fetcher
@@ -729,6 +799,10 @@ module Bundler
       def install_with_build_args(args)
         yield
       end
+
+      def path_separator
+        Gem.path_separator
+      end
     end
 
     # RubyGems 2.1.0
@@ -746,7 +820,13 @@ module Bundler
       end
 
       def backport_ext_builder_monitor
-        require "rubygems/ext"
+        # So we can avoid requiring "rubygems/ext" in its entirety
+        Gem.module_eval <<-RB, __FILE__, __LINE__ + 1
+          module Ext
+          end
+        RB
+
+        require "rubygems/ext/builder"
 
         Gem::Ext::Builder.class_eval do
           unless const_defined?(:CHDIR_MONITOR)
@@ -772,11 +852,27 @@ module Bundler
 
       def use_gemdeps(gemfile)
         ENV["BUNDLE_GEMFILE"] ||= File.expand_path(gemfile)
+        require "bundler/gemdeps"
         runtime = Bundler.setup
         Bundler.ui = nil
         activated_spec_names = runtime.requested_specs.map(&:to_spec).sort_by(&:name)
         [Gemdeps.new(runtime), activated_spec_names]
       end
+
+      if provides?(">= 2.5.2")
+        # RubyGems-generated binstubs call Kernel#gem
+        def binstubs_call_gem?
+          false
+        end
+
+        # only 2.5.2+ has all of the stub methods we want to use, and since this
+        # is a performance optimization _only_,
+        # we'll restrict ourselves to the most
+        # recent RG versions instead of all versions that have stubs
+        def stubs_provide_full_functionality?
+          true
+        end
+      end
     end
   end
 
@@ -795,7 +891,7 @@ module Bundler
       RubygemsIntegration::Transitional.new
     elsif RubygemsIntegration.provides?(">= 1.4.0")
       RubygemsIntegration::Legacy.new
-    else # Rubygems 1.3.6 and 1.3.7
+    else # RubyGems 1.3.6 and 1.3.7
       RubygemsIntegration::Ancient.new
     end
   end
