@@ -157,8 +157,8 @@ module Bundler
         next if gemfiles.empty?
         break false if gemfiles.size == 1
       end
-      if multiple_gemfiles && Bundler.bundler_major_version == 1
-        Bundler::SharedHelpers.major_deprecation 2, \
+      if multiple_gemfiles && Bundler.bundler_major_version == 2
+        Bundler::SharedHelpers.major_deprecation 3, \
           "gems.rb and gems.locked will be preferred to Gemfile and Gemfile.lock."
       end
 
@@ -197,10 +197,12 @@ module Bundler
     def pretty_dependency(dep, print_source = false)
       msg = String.new(dep.name)
       msg << " (#{dep.requirement})" unless dep.requirement == Gem::Requirement.default
+
       if dep.is_a?(Bundler::Dependency)
         platform_string = dep.platforms.join(", ")
         msg << " " << platform_string if !platform_string.empty? && platform_string != Gem::Platform::RUBY
       end
+
       msg << " from the `#{dep.source}` source" if print_source && dep.source
       msg
     end
@@ -223,6 +225,10 @@ module Bundler
       Digest(name)
     end
 
+    def write_to_gemfile(gemfile_path, contents)
+      filesystem_access(gemfile_path) {|g| File.open(g, "w") {|file| file.puts contents } }
+    end
+
   private
 
     def validate_bundle_path
@@ -268,7 +274,15 @@ module Bundler
       until !File.directory?(current) || current == previous
         if ENV["BUNDLE_SPEC_RUN"]
           # avoid stepping above the tmp directory when testing
-          return nil if File.file?(File.join(current, "bundler.gemspec"))
+          gemspec = if ENV["BUNDLE_RUBY"] && ENV["BUNDLE_GEM"]
+            # for Ruby Core
+            "lib/bundler.gemspec"
+          else
+            "bundler.gemspec"
+          end
+
+          # avoid stepping above the tmp directory when testing
+          return nil if File.file?(File.join(current, gemspec))
         end
 
         names.each do |name|
@@ -293,9 +307,16 @@ module Bundler
 
     def set_bundle_variables
       begin
-        Bundler::SharedHelpers.set_env "BUNDLE_BIN_PATH", Bundler.rubygems.bin_path("bundler", "bundle", VERSION)
+        exe_file = Bundler.rubygems.bin_path("bundler", "bundle", VERSION)
+        unless File.exist?(exe_file)
+          exe_file = File.expand_path("../../../exe/bundle", __FILE__)
+        end
+        Bundler::SharedHelpers.set_env "BUNDLE_BIN_PATH", exe_file
       rescue Gem::GemNotFoundException
-        Bundler::SharedHelpers.set_env "BUNDLE_BIN_PATH", File.expand_path("../../../exe/bundle", __FILE__)
+        exe_file = File.expand_path("../../../exe/bundle", __FILE__)
+        # for Ruby core repository
+        exe_file = File.expand_path("../../../../bin/bundle", __FILE__) unless File.exist?(exe_file)
+        Bundler::SharedHelpers.set_env "BUNDLE_BIN_PATH", exe_file
       end
 
       # Set BUNDLE_GEMFILE
